__all__ = ["WebsocketManager"]


import json
from collections.abc import Awaitable, Callable, Sequence
from typing import Any, Literal

from unicex._base import Websocket

from .client import Client

type CallbackType = Callable[[Any], Awaitable[None]]


class WebsocketManager:
    """Менеджер асинхронных вебсокетов для Bitget."""

    _URL: str = "wss://ws.bitget.com/v2/ws/public"
    """Базовый URL для вебсокета."""

    def __init__(self, client: Client | None = None, **ws_kwargs: Any) -> None:
        """Инициализирует менеджер вебсокетов для Bitget.

        Параметры:
            client (`Client | None`): Клиент для выполнения запросов. Нужен, чтобы открыть приватные вебсокеты.
            ws_kwargs (`dict[str, Any]`): Дополнительные аргументы, котоыре прокидываются в `Websocket`.
        """
        self.client = client
        self._ws_kwargs = {"ping_message": "ping", **ws_kwargs}

    def _generate_subscription_message(
        self,
        topic: str,
        market_type: Literal["SPOT", "USDT-FUTURES"],
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> list[str]:
        """Сформировать сообщение для подписки на вебсокет.

        Параметры:
            topic (`str`): Канал подписки (например: "ticker", "candle1m", "depth").
            market_type (`"SPOT" | "USDT-FUTURES"`): Тип рынка для подписки.
            symbol (`str | None`): Торговая пара. Нельзя использовать одновременно с `symbols`.
            symbols (`Sequence[str] | None`): Список торговых пар. Нельзя использовать одновременно с `symbol`.

        Возвращает:
            `str`: JSON-строка с сообщением для подписки на вебсокет.
        """
        if symbol and symbols:
            raise ValueError("Parameters symbol and symbols cannot be used together")
        if not (symbol or symbols):
            raise ValueError("Either symbol or symbols must be provided")

        tickers = [symbol] if symbol else symbols
        streams: list[dict] = [
            {
                "instType": market_type,
                "channel": topic,
                "instId": ticker.upper(),
            }
            for ticker in tickers  # type: ignore
        ]

        return [json.dumps({"op": "subscribe", "args": streams})]

    def trade(
        self,
        callback: CallbackType,
        market_type: Literal["SPOT", "USDT-FUTURES"],
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения сделок.

        https://www.bitget.com/api-doc/spot/websocket/public/Trades-Channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            market_type (`Literal["SPOT", "USDT-FUTURES"]`): Тип рынка.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subsription_messages = self._generate_subscription_message(
            topic="trade",
            market_type=market_type,
            symbol=symbol,
            symbols=symbols,
        )
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=subsription_messages,
            **self._ws_kwargs,
        )

    def ticker(
        self,
        callback: CallbackType,
        market_type: Literal["SPOT", "USDT-FUTURES"],
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения тикеров.

        https://www.bitget.com/api-doc/spot/websocket/public/Tickers-Channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            market_type (`Literal["SPOT", "USDT-FUTURES"]`): Тип рынка.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_messages = self._generate_subscription_message(
            topic="ticker",
            market_type=market_type,
            symbol=symbol,
            symbols=symbols,
        )
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=subscription_messages,
            **self._ws_kwargs,
        )

    def candlestick(
        self,
        callback: CallbackType,
        market_type: Literal["SPOT", "USDT-FUTURES"],
        interval: str,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения данных свечей (candlestick).

        https://www.bitget.com/api-doc/spot/websocket/public/Candlesticks-Channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            market_type (`Literal["SPOT", "USDT-FUTURES"]`): Тип рынка.
            interval (`str`): Интервал свечи, например "1m", "1H", "1Dutc".
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        topic = f"candle{interval}"
        subscription_messages = self._generate_subscription_message(
            topic=topic,
            market_type=market_type,
            symbol=symbol,
            symbols=symbols,
        )
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=subscription_messages,
            **self._ws_kwargs,
        )

    def depth(
        self,
        callback: CallbackType,
        market_type: Literal["SPOT", "USDT-FUTURES"],
        depth_type: str = "books",
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения данных глубины рынка (order book).

        https://www.bitget.com/api-doc/spot/websocket/public/Depth-Channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            market_type (`Literal["SPOT", "USDT-FUTURES"]`): Тип рынка.
            depth_type (`str`): Тип глубины: "books", "books1", "books5", "books15".
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_messages = self._generate_subscription_message(
            topic=depth_type,
            market_type=market_type,
            symbol=symbol,
            symbols=symbols,
        )
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=subscription_messages,
            **self._ws_kwargs,
        )

    def auction(
        self,
        callback: CallbackType,
        symbol: str | None = None,
        symbols: Sequence[str] | None = None,
    ) -> Websocket:
        """Создает вебсокет для получения информации о Call Auction.

        https://www.bitget.com/api-doc/spot/websocket/public/Auction-Channel

        Параметры:
            callback (`CallbackType`): Асинхронная функция обратного вызова для обработки сообщений.
            symbol (`str | None`): Один символ для подписки.
            symbols (`Sequence[str] | None`): Список символов для мультиплекс‑подключения.

        Возвращает:
            `Websocket`: Объект для управления вебсокет соединением.
        """
        subscription_messages = self._generate_subscription_message(
            topic="auction",
            market_type="SPOT",
            symbol=symbol,
            symbols=symbols,
        )
        return Websocket(
            callback=callback,
            url=self._URL,
            subscription_messages=subscription_messages,
            **self._ws_kwargs,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: PublicIncreaseDepthsBatchV3Api.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 6, 31, 1, "", "PublicIncreaseDepthsBatchV3Api.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n$PublicIncreaseDepthsBatchV3Api.proto\x1a\x1fPublicIncreaseDepthsV3Api.proto"^\n\x1ePublicIncreaseDepthsBatchV3Api\x12)\n\x05items\x18\x01 \x03(\x0b\x32\x1a.PublicIncreaseDepthsV3Api\x12\x11\n\teventType\x18\x02 \x01(\tBG\n\x1c\x63om.mxc.push.common.protobufB#PublicIncreaseDepthsBatchV3ApiProtoH\x01P\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "PublicIncreaseDepthsBatchV3Api_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals["DESCRIPTOR"]._loaded_options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = (
        b"\n\034com.mxc.push.common.protobufB#PublicIncreaseDepthsBatchV3ApiProtoH\001P\001"
    )
    _globals["_PUBLICINCREASEDEPTHSBATCHV3API"]._serialized_start = 73
    _globals["_PUBLICINCREASEDEPTHSBATCHV3API"]._serialized_end = 167
# @@protoc_insertion_point(module_scope)



"""Пакет с абстракциями и интерфейсами."""

__all__ = [
    "IUniClient",
    "IUniWebsocketManager",
    "IExchangeInfo",
]

from .exchange_info import IExchangeInfo
from .uni_client import IUniClient
from .uni_websocket_manager import IUniWebsocketManager



__all__ = ["IExchangeInfo"]

import asyncio
import math
from abc import ABC, abstractmethod
from decimal import Decimal
from typing import TYPE_CHECKING, ClassVar

import aiohttp
from loguru import logger

from unicex.enums import MarketType
from unicex.types import TickerInfoItem, TickersInfoDict

if TYPE_CHECKING:
    import loguru


class IExchangeInfo(ABC):
    """Интерфейс для наследников, которые предзагружают и обновляют информацию о бирже."""

    _loaded: bool
    """Флаг, указывающий, была ли информация о бирже загружена."""

    _running: bool
    """Флаг, указывающий, запущена ли фоновая задача для обновления информации о бирже."""

    _tickers_info: TickersInfoDict
    """Словарь с информацией о округлении для каждого тикера."""

    _futures_tickers_info: TickersInfoDict
    """Словарь с информацией о округлении и размере контракта (если есть) для каждого тикера."""

    _logger: "loguru.Logger"
    """Логгер для записи сообщений о работе с биржей."""

    exchange_name: ClassVar[str] = "not_defined_exchange"
    """Название биржи, на которой работает класс."""

    def __init_subclass__(cls, **kwargs):
        """Инициализация подкласса. Функция нужна, чтобы у каждого наследника была своя копия атрибутов."""
        super().__init_subclass__(**kwargs)
        cls._tickers_info = {}
        cls._loaded = False
        cls._running = False
        cls._logger = logger

    @classmethod
    async def start(cls, update_interval_seconds: int = 60 * 60) -> None:
        """Запускает фоновую задачу с бесконечным циклом для загрузки данных."""
        cls._running = True
        asyncio.create_task(cls._load_exchange_info_loop(update_interval_seconds))

    @classmethod
    async def stop(cls) -> None:
        """Останавливает фоновую задачу для обновления информации о бирже."""
        cls._running = False

    @classmethod
    async def set_logger(cls, logger: "loguru.Logger") -> None:
        """Устанавливает логгер для записи сообщений о работе с биржей."""
        cls._logger = logger

    @classmethod
    async def _load_exchange_info_loop(cls, update_interval_seconds: int) -> None:
        """Запускает бесконечный цикл для загрузки данных о бирже."""
        while cls._running:
            try:
                await cls.load_exchange_info()
            except Exception as e:
                cls._logger.error(f"Error loading exchange data for {cls.exchange_name}: {e}")
            for _ in range(update_interval_seconds):
                if not cls._running:
                    break
                await asyncio.sleep(1)

    @classmethod
    async def load_exchange_info(cls) -> None:
        """Принудительно вызывает загрузку информации о бирже."""
        async with aiohttp.ClientSession() as session:
            try:
                await cls._load_spot_exchange_info(session)
                cls._logger.debug(f"Loaded spot exchange data for {cls.exchange_name} ")
            except Exception as e:
                cls._logger.error(f"Error loading spot exchange data for {cls.exchange_name}: {e}")
            try:
                await cls._load_futures_exchange_info(session)
                cls._logger.debug(f"Loaded futures exchange data for {cls.exchange_name} ")
            except Exception as e:
                cls._logger.error(
                    f"Error loading futures exchange data for {cls.exchange_name}: {e}"
                )
        cls._loaded = True

    @classmethod
    @abstractmethod
    async def _load_spot_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        """Загружает информацию о бирже для спотового рынка."""
        ...

    @classmethod
    @abstractmethod
    async def _load_futures_exchange_info(cls, session: aiohttp.ClientSession) -> None:
        """Загружает информацию о бирже для фьючерсного рынка."""
        ...

    @classmethod
    def get_ticker_info(
        cls, symbol: str, market_type: MarketType = MarketType.SPOT
    ) -> TickerInfoItem:  # type: ignore[reportReturnType]
        """Возвращает информацию о тикере по его символу."""
        try:
            if market_type == MarketType.SPOT:
                return cls._tickers_info[symbol]
            return cls._futures_tickers_info[symbol]
        except KeyError as e:
            cls._handle_key_error(e, symbol)

    @classmethod
    def get_futures_ticker_info(cls, symbol: str) -> TickerInfoItem:
        """Возвращает информацию о тикере фьючерсов по его символу."""
        return cls.get_ticker_info(symbol, MarketType.FUTURES)

    @classmethod
    def round_price(
        cls, symbol: str, price: float, market_type: MarketType = MarketType.SPOT
    ) -> float:  # type: ignore
        """Округляет цену до ближайшего возможного значения."""
        try:
            if market_type == MarketType.SPOT:
                precision = cls._tickers_info[symbol]["tick_precision"]
                step = cls._tickers_info[symbol]["tick_step"]
            else:
                precision = cls._futures_tickers_info[symbol]["tick_precision"]
                step = cls._futures_tickers_info[symbol]["tick_step"]
            if precision:
                return cls._floor_round(price, precision)
            return cls._floor_to_step(price, step)  # type: ignore
        except KeyError as e:
            cls._handle_key_error(e, symbol)

    @classmethod
    def round_quantity(
        cls, symbol: str, quantity: float, market_type: MarketType = MarketType.SPOT
    ) -> float:  # type: ignore
        """Округляет объем до ближайшего возможного значения."""
        try:
            if market_type == MarketType.SPOT:
                precision = cls._tickers_info[symbol]["size_precision"]
                step = cls._tickers_info[symbol]["size_step"]
            else:
                precision = cls._futures_tickers_info[symbol]["size_precision"]
                step = cls._futures_tickers_info[symbol]["size_step"]
            if precision:
                return cls._floor_round(quantity, precision)
            return cls._floor_to_step(quantity, step)  # type: ignore
        except KeyError as e:
            cls._handle_key_error(e, symbol)

    @classmethod
    def round_futures_price(cls, symbol: str, price: float) -> float:
        """Округляет цену до ближайшего возможного значения на фьючерсах."""
        return cls.round_price(symbol, price, MarketType.FUTURES)

    @classmethod
    def round_futures_quantity(cls, symbol: str, quantity: float) -> float:
        """Округляет объем до ближайшего возможного значения на фьючерсах."""
        return cls.round_quantity(symbol, quantity, MarketType.FUTURES)

    @staticmethod
    def _floor_to_step(value: float, step: float) -> float:
        """Округляет число вниз до ближайшего кратного шага.

        Принимает:
            value (float): Исходное число.
            step: (float): Шаг округления (> 0).

        Возвращает:
            Число, округлённое вниз до кратного step.

        Примеры:
            >>> floor_to_step(0.16, 0.05)
            0.15
            >>> floor_to_step(16, 5)
            15
            >>> floor_to_step(1.2345, 0.01)
            1.23
            >>> floor_to_step(-1.23, 0.1)
            -1.3
            >>> floor_to_step(100, 25)
            100

        """
        if step <= 0:
            raise ValueError("step must be > 0")
        result = math.floor(value / step) * step
        digits = abs(Decimal(str(step)).as_tuple().exponent)  # type: ignore
        return round(result, digits)

    @staticmethod
    def _floor_round(value: float, digits: int) -> float:
        """Округляет число вниз до указанного количества знаков после запятой."""
        factor = 10**digits
        return math.floor(value * factor) / factor

    @classmethod
    def _handle_key_error(cls, exception: KeyError, symbol: str) -> None:
        """Обрабатывает KeyError при получении информации о тикере."""
        cls._check_loaded()
        raise KeyError(f"Symbol {symbol} not found") from exception

    @classmethod
    def _check_loaded(cls) -> None:
        """Проверяет, загружены ли данные об обмене."""
        if not cls._loaded:
            raise ValueError("Exchange data not loaded") from None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: PublicMiniTickersV3Api.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 6, 31, 1, "", "PublicMiniTickersV3Api.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1cPublicMiniTickersV3Api.proto\x1a\x1bPublicMiniTickerV3Api.proto"?\n\x16PublicMiniTickersV3Api\x12%\n\x05items\x18\x01 \x03(\x0b\x32\x16.PublicMiniTickerV3ApiB?\n\x1c\x63om.mxc.push.common.protobufB\x1bPublicMiniTickersV3ApiProtoH\x01P\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "PublicMiniTickersV3Api_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals["DESCRIPTOR"]._loaded_options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = (
        b"\n\034com.mxc.push.common.protobufB\033PublicMiniTickersV3ApiProtoH\001P\001"
    )
    _globals["_PUBLICMINITICKERSV3API"]._serialized_start = 61
    _globals["_PUBLICMINITICKERSV3API"]._serialized_end = 124
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: PublicMiniTickerV3Api.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 6, 31, 1, "", "PublicMiniTickerV3Api.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1bPublicMiniTickerV3Api.proto"\xf4\x01\n\x15PublicMiniTickerV3Api\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\r\n\x05price\x18\x02 \x01(\t\x12\x0c\n\x04rate\x18\x03 \x01(\t\x12\x11\n\tzonedRate\x18\x04 \x01(\t\x12\x0c\n\x04high\x18\x05 \x01(\t\x12\x0b\n\x03low\x18\x06 \x01(\t\x12\x0e\n\x06volume\x18\x07 \x01(\t\x12\x10\n\x08quantity\x18\x08 \x01(\t\x12\x15\n\rlastCloseRate\x18\t \x01(\t\x12\x1a\n\x12lastCloseZonedRate\x18\n \x01(\t\x12\x15\n\rlastCloseHigh\x18\x0b \x01(\t\x12\x14\n\x0clastCloseLow\x18\x0c \x01(\tB>\n\x1c\x63om.mxc.push.common.protobufB\x1aPublicMiniTickerV3ApiProtoH\x01P\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "PublicMiniTickerV3Api_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals["DESCRIPTOR"]._loaded_options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = (
        b"\n\034com.mxc.push.common.protobufB\032PublicMiniTickerV3ApiProtoH\001P\001"
    )
    _globals["_PUBLICMINITICKERV3API"]._serialized_start = 32
    _globals["_PUBLICMINITICKERV3API"]._serialized_end = 276
# @@protoc_insertion_point(module_scope)



from typing import Any

from unicex.types import (
    AggTradeDict,
    KlineDict,
    OpenInterestDict,
    OpenInterestItem,
    TickerDailyDict,
    TickerDailyItem,
    TradeDict,
)
from unicex.utils import catch_adapter_errors, decorate_all_methods


@decorate_all_methods(catch_adapter_errors)
class Adapter:
    """Адаптер для унификации данных с Bitget API."""

    @staticmethod
    def tickers(raw_data: Any, only_usdt: bool) -> list[str]:
        """Преобразует сырой ответ, в котором содержатся данные о тикерах в список тикеров.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.
            only_usdt (bool): Флаг, указывающий, нужно ли включать только тикеры в паре к USDT.

        Возвращает:
            list[str]: Список тикеров.
        """
        return [
            item["symbol"]
            for item in raw_data["data"]
            if item["symbol"].endswith("USDT") or not only_usdt
        ]

    @staticmethod
    def ticker_24hr(raw_data: Any) -> TickerDailyDict:
        """Преобразует сырой ответ, в котором содержатся данные о тикере за последние 24 часа
        в унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            TickerDailyDict: Словарь, где ключ - тикер, а значение - статистика за последние 24 часа.
        """
        return {
            item["symbol"]: TickerDailyItem(
                p=round(float(item["change24h"]) * 100, 2),  # конвертируем в проценты
                v=float(item["baseVolume"]),  # объём в COIN
                q=float(item["usdtVolume"]),  # объём в USDT
            )
            for item in raw_data["data"]
        }

    @staticmethod
    def last_price(raw_data: Any) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о последних ценах тикеров
        в унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - последняя цена.
        """
        return {item["symbol"]: float(item["lastPr"]) for item in raw_data["data"]}

    @staticmethod
    def klines(raw_data: Any, symbol: str) -> list[KlineDict]:
        """Преобразует сырой ответ, в котором содержатся данные о котировках тикеров в
        унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.
            symbol (str): Тикер, для которого нужно преобразовать данные.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        return [
            KlineDict(
                s=symbol,
                t=int(kline[0]),
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[5]),
                q=float(kline[6]),
                T=None,
                x=None,
            )
            for kline in sorted(
                raw_data["data"],
                key=lambda x: int(x[0]),  # Bitget присылает пачку трейдов в обратном порядке
            )
        ]

    @staticmethod
    def funding_rate(raw_data: Any) -> dict[str, float]:
        """Преобразует сырой ответ, в котором содержатся данные о ставках финансирования
        тикеров в унифицированный формат.

        Параметры:
            raw_data (Any): Сырой ответ с биржи.

        Возвращает:
            dict[str, float]: Словарь, где ключ - тикер, а значение - ставка финансирования.
        """
        return {item["symbol"]: float(item["fundingRate"]) * 100 for item in raw_data["data"]}

    @staticmethod
    def klines_message(raw_msg: Any) -> list[KlineDict]:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        свече/свечах в унифицированный вид.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[KlineDict]: Список словарей, где каждый словарь содержит данные о свече.
        """
        symbol = raw_msg["arg"]["instId"]

        return [
            KlineDict(
                s=symbol,
                t=int(kline[0]),
                o=float(kline[1]),
                h=float(kline[2]),
                l=float(kline[3]),
                c=float(kline[4]),
                v=float(kline[5]),
                q=float(kline[6]),
                T=None,
                x=None,
            )
            for kline in sorted(
                raw_msg["data"],
                key=lambda x: int(x[0]),  # Bitget присылает пачку трейдов в обратном порядке
            )
        ]

    @staticmethod
    def trades_message(raw_msg: Any) -> list[TradeDict]:
        """Преобразует сырое сообщение вебсокета со сделками в унифицированный формат.

        Параметры:
            raw_msg (Any): Сырое сообщение с вебсокета.

        Возвращает:
            list[TradeDict]: Список сделок в унифицированном формате.
        """
        symbol = raw_msg["arg"]["instId"]

        return [
            AggTradeDict(
                t=int(trade["ts"]),
                s=symbol,
                S=trade["side"].upper(),
                p=float(trade["price"]),
                v=float(trade["size"]),
            )
            for trade in sorted(
                raw_msg["data"],
                key=lambda x: int(x["ts"]),  # Bitget присылает пачку трейдов в обратном порядке
            )
        ]

    @staticmethod
    def open_interest(raw_data: Any) -> OpenInterestDict:
        """Преобразует сырое сообщение с вебсокета, в котором содержится информация о
        объеме открытых позиций в унифицированный вид.

        Параметры:
            raw_data (Any): Сырое сообщение с вебсокета.

        Возвращает:
            `OpenInterestDict`: Cловарь, где ключи - название тикера, а значения - объемы открытых позиций в монетах.
        """
        return {
            i["symbol"]: OpenInterestItem(
                t=int(i["ts"]),
                v=float(i["holdingAmount"]),
            )
            for i in raw_data["data"]
        }
